/*
   GRAFIK.C

   simple graphic functions not depending on hardware
   and based on gr_prim.c 

   history:
   8.7.95   Till  first version
  10.7.95   Till  added font routines
  28.4.96   Till  changed font routines for use with metafont
  13.11.96  Chip  display of title-pic changed
  14.11.96  Chip  put_bitmap nur durchsichtig
  14.11.96  Chip  korrekte Farben fuer Monochrom-Schirme
  26.11.96  Chip  put_bitmap mit 16 Bit, keine Multicolor-Fonts mehr
  27.11.96  Chip  Typen auf tm_* umgestellt
  19.03.97  Chip  sauberes Ende bei Absturz versucht
   6.01.98  Till  erweitert fuer Textur und Unterstrich im Zeichensatz
  17.01.98  Till  Sachen fuer Statusleiste und Spielerfarben
*/

#include <stdlib.h>

#include "ctypes.h"
#include "grafik.h"
#include "console.h"
#include "gr_prim.h"
#ifdef TEXTURE
#include "texture.h"
#endif
#include "net.h"

tm_uword width,height,colors, pf_height;

#ifdef TEXTURE
tm_bool use_texture=FALSE;
#endif

#ifdef USE_MALLOC
tm_void *compass_buffer;
#endif

tm_bool init_screen(tm_uword argc, tm_char **argv)
{
  if(!open_screen(argc,argv,&width,&height,&colors))
    return(ERROR);

  pf_height = height-(2*font_height);
  save_colors();

  return(OK);
}

tm_void exit_screen(tm_void)
{
#ifdef USE_MALLOC
  free_rect(compass_buffer);
#endif

  restore_colors();
  close_screen();
}

tm_void print_message(tm_bool line, tm_char *msg, tm_word color) {
  static tm_word old_width[2]={-1,-1};

  if(old_width[line&1]==-1) old_width[line&1]=width;

  draw_rect(0,pf_height+(line?font_height:0),old_width[line&1],font_height,(colors<16)?BW_BLACK:COL_BLACK);
  draw_text(0,pf_height+(line?font_height:0),msg,(colors<16)?BW_WHITE:color);
  old_width[line&1] = get_text_width(msg);

  /* Statuszeile in zweiten Puffer kopieren */
  update_double_buffer();
}

#define UNDERSCORE '-'  /* 123/124 */    /* character to use as underscore */

/* Kompassdaten */
#define ADD_Y      (pf_height+font_height)
#define ADD_X      (width-font_height-2)
#define DISC_RAD   (font_height-1)
tm_uword border_col,inner_col,needle_col;

tm_void draw_compass(tm_ushort rot) {
  tm_word size=32768l/(font_height-1);
  tm_word r1,r2,r3,r4;

#ifdef USE_MALLOC
  static tm_bool compass_buffered=FALSE;
#endif

  r1 = rot >> 4;
  r2 = ((rot+0x8000-1000)&0xffff) >> 4;
  r3 = ((rot+0x8000+1000)&0xffff) >> 4;
  r4 = ((rot+0x8000)&0xffff) >> 4;

#ifdef USE_MALLOC
  if(!compass_buffered) {
#endif
    /* Kompass zeichnen */
    draw_disc(ADD_X,ADD_Y,DISC_RAD,inner_col,border_col);
#ifdef USE_MALLOC
    compass_buffer = save_rect(width-2*font_height-2, pf_height, 2*font_height, 2*font_height);
    compass_buffered=TRUE;
  } else
    restore_rect(compass_buffer);
#endif

  draw_line(ADD_X-(sintab[r1]/size), ADD_Y-(costab[r1]/size),
	    ADD_X-(sintab[r2]/size), ADD_Y-(costab[r2]/size),needle_col);
  draw_line(ADD_X-(sintab[r1]/size), ADD_Y-(costab[r1]/size),
	    ADD_X-(sintab[r3]/size), ADD_Y-(costab[r3]/size),needle_col);
  draw_line(ADD_X-(sintab[r1]/size), ADD_Y-(costab[r1]/size),
	    ADD_X-(sintab[r4]/size), ADD_Y-(costab[r4]/size),needle_col);
}

/* simply draw text */
tm_void draw_text(tm_uword x, tm_uword y,tm_char *string,tm_uword color)
{
  tm_uword i=0;
  tm_bool  esc=FALSE,uml=FALSE,killnext=FALSE;

  while(string[i]!=0)
  {
    if((!esc)&&(string[i]=='\\')) {  /* Escape character */
      esc = TRUE;
      uml = string[i+1];
      if((uml == '\"')&&(string[i+2]=='s')) killnext=TRUE;
      i+=2;
    } else {

      if(charTable[string[i]].width!=0) { /* Buchstabe existiert */
	if(!killnext) {
	  if(string[i]=='_')
	    put_bitmap(x,y+charTable['M'].offset+charTable['M'].height,
		       charTable[UNDERSCORE].width,
		       charTable[UNDERSCORE].height,charTable[UNDERSCORE].data,
		       color);
	  else
	    put_bitmap(x,y+charTable[string[i]].offset,charTable[string[i]].width,
		       charTable[string[i]].height,charTable[string[i]].data,
		       color);
	}

	if(uml=='\"') {  /* german umlaut */
	  if(string[i]=='s') {
	    put_bitmap(x,y+charTable[0x19].offset,charTable[0x19].width,charTable[0x19].height,
		       charTable[0x19].data,color);
	  } else { /* put double point on top of char */
	    put_bitmap(x+(charTable[string[i]].width-charTable[127].width)/2,
		       y+charTable[string[i]].offset-charTable[127].height*3/2,
		       charTable[127].width,charTable[127].height,charTable[127].data,
		       color);
	  }
	}
      }

      esc = FALSE;
      uml = FALSE;
      killnext=FALSE;

      if(string[i]=='_')
	x+=charTable[UNDERSCORE].step;
      else
	x+=charTable[string[i]].step;

      i++;
    }
  }
}

/* draw string with specific width by altering the space */
/* betweeen the words */
tm_void draw_text_width(tm_uword x, tm_uword y, tm_char *string,
                        tm_uword color, tm_uword width)
{
  tm_uword i,j,real_width,spaces,space_width;
  tm_char  buffer[100];

  /* get the length over all words excluding spaces */
  i=0; j=0; real_width=0; spaces=0;
  do {
    if((string[i]==' ')||(string[i]==0)) {
      if(string[i]==' ') spaces++;
      buffer[j]=0;
      real_width += get_text_width(buffer);
      j=0;
    } else
      buffer[j++]=string[i];
  }
  while(string[i++]!=0);

  /* this only works with at least one space */
  if(spaces>0) {
    space_width = (width-real_width)/spaces;

    /* and now: do the text */
    i=0; j=0; spaces=0;
    do {
      if((string[i]==' ')||(string[i]==0)) {
	buffer[j]=0;
	draw_text(x+spaces, y, buffer, color);
	spaces += get_text_width(buffer)+space_width;
	j=0;
      } else
	buffer[j++]=string[i];
    }
    while(string[i++]!=0);
  }
  else  /* without spaces there's not very much to do */
    draw_text(x, y, string, color);
}

/* return width of a str in pixels */ 
tm_uword get_text_width(tm_char *str)
{
  tm_uword i=0,width=0;
  tm_bool  esc=FALSE,uml=FALSE,killnext=FALSE;
  
  while(str[i]!=0) {
    if((!esc)&&(str[i]=='\\')) {  /* Escape character */
      esc = TRUE;
      uml = str[i+1];
      if((uml == '\"')&&(str[i+2]=='s')) killnext=TRUE;
      i+=2;
    } else {
      if(!killnext) {
	if(str[i] == '_')
	  width += charTable[UNDERSCORE].step;
	else
	  width += charTable[str[i]].step;
      } else
	if(uml=='\"')   /* german umlaut */
	  if(str[i]=='s')
	    width+=charTable[0x19].step;

      i++;
      esc = FALSE;
      uml = FALSE;
      killnext=FALSE;
    }
  }
  return(width);
}
