/*
   snd2c.c                                     (c) 1997 Till Harbaum

   converts a bunch of 8kHz/ulaw sun style .au files into a c source
   file converting the files to a particular sample rate and format

   ulaw2linear code stolen from sox

   example:
   snd2c sounds.c raw8 22050 boing.au shot.au

*/

#include <stdio.h>
#include "machine.h"

#define MAX_SOUNDS 64

#ifdef LITTLE_ENDIAN
#define CONV(a)  (((a&0xff000000)>>24)|((a&0xff0000)>>8)|((a&0xff00)<<8)|((a&0xff)<<24))
#else
#define CONV(a)  (a)
#endif

typedef enum {ulaw,raw8} snd_format;

typedef struct {
  unsigned signature;       /* .snd */
  unsigned hlen;
  unsigned dlen;
  unsigned a;               /* ulaw = 1 */
  unsigned rate;            /* ulaw = 8000 */
  unsigned b;
} au_hdr;

#define SIGN_BIT        (0x80)          /* Sign bit for a A-law byte. */
#define QUANT_MASK      (0xf)           /* Quantization field mask. */
#define NSEGS           (8)             /* Number of A-law segments. */
#define SEG_SHIFT       (4)             /* Left shift for segment number. */
#define SEG_MASK        (0x70)          /* Segment field mask. */
#define BIAS            (0x84)          /* Bias for linear code. */

int length[MAX_SOUNDS];

int
ulaw2linear(unsigned char u_val) {
  int t;

  /* Complement to obtain normal u-law value. */
  u_val = ~u_val;

  t = ((u_val & QUANT_MASK) << 3) + BIAS;
  t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
  
  return ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));
}

int
main(int argc, char *argv[]) {

  FILE *out,*in;
  int  freq,n,m,snds=0;
  snd_format form;
  au_hdr hdr;
  char *mem, *memx;

  if(argc<5) {
    printf("Usage: snd2c fname format frequency file [more files]\n");
    printf("       supported formats: ulaw raw8\n");
    exit(1);
  }

  if ((strcmp(argv[2],"ulaw")!=0)&&
      (strcmp(argv[2],"raw8")!=0)) {
    printf("       supported formats: ulaw raw8\n");
    exit(1);
  }

  if(strcmp(argv[2],"ulaw")==0) form=ulaw;
  if(strcmp(argv[2],"raw8")==0) form=raw8;
     
  if((out=fopen(argv[1],"wb"))==NULL) {
    printf("unable to open output file '%s'\n",argv[1]);
    exit(1);
  }

  fprintf(out,"/* sounds for TheMaze */\n");

  freq = atoi(argv[3]);
  if(freq == 0) {
    fprintf(out,"/* this file is intentionally left blank */\n"
	        "/* (due to lack of sound support for this system) */\n");
    fclose(out);
    exit(0);
  }

  fprintf(out,"#include \"ctypes.h\"\n");

  for(n=0;n<(argc-4);n++) {
    printf("adding '%s' ...\n",argv[n+4]);

    if((in=fopen(argv[n+4],"rb"))!=NULL) {
      fread(&hdr,sizeof(au_hdr),1,in);
      fseek(in,CONV(hdr.hlen),0);      /* seek behind header  */

      if(CONV(hdr.rate) == freq) {
	mem=(char*)malloc(CONV(hdr.dlen));      /* get some memory ... */
	fread(mem,CONV(hdr.dlen),1,in);  /* read rest of file   */
      } else {
	printf("scaling %d%%\n",100*freq/CONV(hdr.rate));
	mem=(char*)malloc(CONV(hdr.dlen)*freq/CONV(hdr.rate));  /* get some memory ... */
	memx=(char*)malloc(CONV(hdr.dlen));               /* get more memory ... */
	fread(memx,CONV(hdr.dlen),1,in);                  /* read rest of file   */
	for(m=0;m<CONV(hdr.dlen)*freq/CONV(hdr.rate);m++)       /* rescan */
	  mem[m]=memx[m*CONV(hdr.rate)/freq];
	free(memx);
      }
      fclose(in);

      /* convert to linear? */
      if(form == raw8) 
	for(m=0;m<CONV(hdr.dlen)*freq/CONV(hdr.rate);m++)
	  mem[m]=ulaw2linear(mem[m])>>8;

      length[snds]=CONV(hdr.dlen)*freq/CONV(hdr.rate);

      fprintf(out,"tm_uchar tm_sound_%d[]={\n\t",snds);
      for(m=0;m<CONV(hdr.dlen)*freq/CONV(hdr.rate);m++)
	fprintf(out,"0x%02x%s%s",((int)mem[m])&0xff,
		(m!=CONV(hdr.dlen)*freq/CONV(hdr.rate)-1)?",":"",
		(m%14==13)?"\n\t":"");	
      fprintf(out,"};\n\n");

      snds++;
      free(mem);
    }
  }

  fprintf(out,"tm_char *tm_sounds[]={");
  for(m=0;m<snds;m++)
    fprintf(out,"tm_sound_%d%s",m,(m!=snds-1)?",":"");
  fprintf(out,"};\n\ntm_word tm_sound_len[]={");
  for(m=0;m<snds;m++)
    fprintf(out,"%d%s",length[m],(m!=snds-1)?",":"");
  fprintf(out,"};\n/* --- EOF --- */");

  fclose(out);

  return(0);
}
