/*
  lpj.c - line printer joystick driver 
          for C64 style joysticks under Linux

  (c) 1998 Till Harbaum

  INSTALL:

*/

#include <linux/config.h>
#include <unistd.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/ioport.h>  /* check/request/release_region */
#include <linux/delay.h>
#include <asm/io.h>
#include "lpj.h"

#define LPJ_MAJOR  122
#define LPJ_IOSIZE   4

#define noDEBUG

#define LPJ_MAX_JOYSTICKS 3
const int lpj_iobase_table[LPJ_MAX_JOYSTICKS]={ 0x378, 0x278, 0x3bc};  /* printer base addresses to test */
int lpj_iobase[LPJ_MAX_JOYSTICKS];

static int 
lpj_read(struct inode * inode, struct file * file, char *adr, int len) {
  unsigned int dev = MINOR(inode->i_rdev), port;
  unsigned char joy;
#ifdef DEBUG
  printk("lpj: lpj_read%d/%d\n", dev, len);
#endif

  if(len!=1)
    return -EINVAL;

  if(lpj_iobase[dev]<0)
    return -EIO;

  port = inb_p(lpj_iobase[dev]+1);
  joy=0;
  if(!(port&0x08))   joy|=LPJ_JOY_FIRE;
  if(!(port&0x10))   joy|=LPJ_JOY_UP;
  if(!(port&0x20))   joy|=LPJ_JOY_DOWN;
  if(!(port&0x40))   joy|=LPJ_JOY_RIGHT;
  if(port&0x80)      joy|=LPJ_JOY_LEFT;

  if(!(verify_area(VERIFY_WRITE, adr, 1))) {
    memcpy_tofs(adr,&joy,1);
  } else
    return -EINVAL;
  
  return 1;
}

static int 
lpj_open(struct inode * inode, struct file * file)
{
  unsigned int dev = MINOR(inode->i_rdev);
#ifdef DEBUG
  printk("lpj: lpj_open%d\n", dev);
#endif

  if(dev>=LPJ_MAX_JOYSTICKS)
    return -EIO;

  if(lpj_iobase[dev]>=0)
    return -EBUSY;

  if (check_region(lpj_iobase_table[dev], LPJ_IOSIZE) < 0) {
    printk("lpj: unable to allocate io space at 0x%03x\n",
	   lpj_iobase_table[dev]);
    return -EIO;
  }

  request_region(lpj_iobase_table[dev], LPJ_IOSIZE, "lp joystick");
  lpj_iobase[dev] = lpj_iobase_table[dev];

  MOD_INC_USE_COUNT;
  return 0;
}

static void 
lpj_release(struct inode * inode, struct file * file)
{
  unsigned int dev = MINOR(inode->i_rdev);
#ifdef DEBUG
  printk("lpj: lpj_release%d\n", dev);
#endif

  if(lpj_iobase[dev]<0)
    return;

  release_region(lpj_iobase[dev], LPJ_IOSIZE);
  lpj_iobase[dev]=-1;
 
  MOD_DEC_USE_COUNT;
}


static struct file_operations lpj_fops = {
  NULL,           /* lpj_seek */
  lpj_read,       /* lpj_read */
  NULL,           /* lpj_write */
  NULL,           /* lpj_readdir */
  NULL,           /* lpj_select */
  NULL,           /* lpj_ioctl */
  NULL,           /* lpj_mmap */
  lpj_open,       /* lpj_open */
  lpj_release,    /* lpj_release */
};

int 
init_module(void) {
  int i;

#ifdef DEBUG
  printk("lpj: init_module\n");
#endif

  if (register_chrdev(LPJ_MAJOR,"lp joystick",&lpj_fops)) {
    printk("fhipps vampire: unable to get major %d\n", LPJ_MAJOR);
    return -EIO;
  }

  for(i=0;i<LPJ_MAX_JOYSTICKS;i++)
    lpj_iobase[i]=-1;

  return(0);
}

/* MOD_INC_USE_COUNT; */

void cleanup_module(void)
{
  int i;

#ifdef DEBUG
  printk("lpj: cleanup & unregister\n");
#endif

  for(i=0;i<LPJ_MAX_JOYSTICKS;i++)
    if(lpj_iobase[i]>=0)
      release_region(lpj_iobase[i], LPJ_IOSIZE);

  unregister_chrdev(LPJ_MAJOR,"lp joystick");
}


